//==========================================================================
// File: facetree.c
// Description: An implementation of Kranakis, Singh and Urrutia routing
// Author: Pat Morin
// Created: 
//==========================================================================
#include <LEDA/segment.h>

#include "facetree.h"
#include "util.h"


// Use this define to get a debug window that pops up whenever this
// function is called.
// #define DBGWIN

#ifdef DBGWIN
#include <LEDA/window.h>
#endif DBGWIN

// difference types of entries into a face
enum {E_VERTEX, E_EDGE, E_NONE};

//==========================================================================
// Find the FACETREE path from s to t
// Complexity: O(n)
//========================================================================
void FACETREE_PATH(const graph &G, const node_array<POINT> &pos, 
		   node s, node t, list<edge> &P)
{
  edge e;
#ifdef DBGWIN
  window W(300+10, 300+20, "unit graph debug window");  
  W.button("dismiss",1);
  W.open(window::center,window::center);
  W.init(-.1, 1.2, -.1);
  W.set_node_width(3);
  node v;
  forall_nodes (v, G) {
    W.draw_filled_node(pos[v], blue);
  }
  forall_edges (e, G) {
    W.draw_segment(pos[G.source(e)], pos[G.target(e)], black);
  }
  W.draw_filled_node(pos[s], red);
  W.draw_filled_node(pos[t], green);
  W.set_line_width(3);
  W.set_mode(xor_mode);
#endif // DBGWIN

  P.clear();

  // first check if s==t or s is adjacent to t
  if (s == t) return;
  forall_adj_edges (e, s) { 
    if (G.target(e) == t) { // we're done!
      P.append(e);
      return;
    }
  }
  
  // find an edge of the first face that intersects SEGMENT (s,t)
  e = find_first(G, pos, s, t);

  // start walking
  while (1 < 2) {
    // Find the entry edge/vertex of the current face
    edge start = e, entry = nil, eold = nil;
    point pt = pos[G.source(e)];
    int type = E_NONE;
    int count = 0, entry_count = 0;
    do {
#ifdef DBGWIN
      W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
      if (eold != nil)
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
#endif // DBGWIN
      P.append(e);
      if (G.target(e) == t) {
	if (!valid_path(G, P, s, t)) {
	  cerr << "FACETREE_PATH: returning an invalid path" << endl;
	  forall (e, P) {
	    cerr << pos[G.source(e)] << " " << pos[G.target(e)] << endl;
	  }
	  P.clear();
	}
	return;
      }
      count++;
      POINT px;
      SEGMENT se(pos[G.source(e)], pos[G.target(e)]);
      int dtype = robust_distance_test(se, pos[t], px);
      if (dtype == D_SOURCE) {
	px = POINT(HUGE_VAL, HUGE_VAL);
      }
      if (px.sqr_dist(pos[t]) < pt.sqr_dist(pos[t])
	  || (px.sqr_dist(pos[t]) == pt.sqr_dist(pos[t])
	      && type == E_EDGE && dtype == D_TARGET)) {
	type = (dtype == D_TARGET) ? E_VERTEX : E_EDGE;
	pt = px;
	entry = e;
      }
      eold = e;
      e = G.face_cycle_succ(e);
    } while (e != start);

    // sanity check
    if (entry == nil) {
      cerr << "FACETREE_PATH: ventry == nil && entry == nil" << endl;
      cout << pos[s] << endl << pos[t] << endl;
      P.clear();
      return;
    }

    // return to the entry edge/vertex and cross over into the next face
    if (entry_count > count / 2) {
      e = G.face_cycle_pred(e);
      while (e != entry) {
	P.append(G.reversal(e));
	eold = e;
	e = G.face_cycle_pred(e);
#ifdef DBGWIN
	W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
#endif // DBGWIN
      }
      if (type == E_EDGE) {
	e = G.reversal(e);
      } else {
	e = find_first(G, pos, G.target(e), t);
      }
    } else {
      while (e != entry) {
#ifdef DBGWIN
	W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
#endif // DBGWIN
	P.append(e);
	eold = e;
	e = G.face_cycle_succ(e);
      }
      if (type == E_EDGE) {
	e = G.cyclic_adj_pred(e);
      } else {
	P.append(e);
	e = find_first(G, pos, G.target(e), t);
      }
    }
  }
}



