//==========================================================================
// File: ksu.c
// Description: An implementation of Kranakis, Singh and Urrutia routing
// Author: Pat Morin
// Created: 
//==========================================================================
#include <LEDA/segment.h>

#include "ksu.h"
#include "util.h"


// Use this define to get a debug window that pops up whenever this
// function is called.
// #define DBGWIN

#ifdef DBGWIN
#include <LEDA/window.h>
#endif DBGWIN

//==========================================================================
// Find the KSU path from s to t
// Complexity: O(n)
//========================================================================
void KSU_PATH(const graph &G, const node_array<POINT> &pos, 
	      node s, node t, list<edge> &P)
{
  edge e;
#ifdef DBGWIN
  window W(300+10, 300+20, "unit graph debug window");  
  W.button("dismiss",1);
  W.open(window::center,window::center);
  W.init(-.1, 1.2, -.1);
  W.set_node_width(3);
  node v;
  forall_nodes (v, G) {
    W.draw_filled_node(pos[v], blue);
  }
  forall_edges (e, G) {
    W.draw_segment(pos[G.source(e)], pos[G.target(e)], black);
  }
  W.draw_filled_node(pos[s], red);
  W.draw_filled_node(pos[t], green);
  W.set_line_width(3);
  W.set_mode(xor_mode);
#endif // DBGWIN

  P.clear();

  // first check if s==t or s is adjacent to t
  if (s == t) return;
  forall_adj_edges (e, s) { 
    if (G.target(e) == t) { // we're done!
      P.append(e);
      return;
    }
  }
  
  // find an edge of the first face that intersects SEGMENT (s,t)
  e = find_first(G, pos, s, t);

  // always refer to the original segment (s,t) to avoid cumulative
  // arithmetic errors
  SEGMENT st(pos[s], pos[t]);

  // start walking
  POINT p = pos[s];
  while (1 < 2) {
    // Find the entry edge of the current face
    edge start = e, entry = nil, eold = nil;
    int count = 0, entry_count = 0;
    do {
#ifdef DBGWIN
      W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
      if (eold != nil)
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
#endif // DBGWIN
      P.append(e);
      if (G.target(e) == t) {
	if (!valid_path(G, P, s, t)) {
	  cerr << "KSU_PATH: returning an invalid path" << endl;
	  forall (e, P) {
	    cerr << pos[G.source(e)] << " " << pos[G.target(e)] << endl;
	  }
	  P.clear();
	}
	return;
      }
      count++;
      POINT px;
      SEGMENT se(pos[G.source(e)], pos[G.target(e)]);
      int type = robust_intersection_test(st, se, px);
      if (type == I_NONE || type == I_SOURCE) {
	px = POINT(HUGE_VAL, HUGE_VAL);
      }
      if (type == I_OVERLAP) {
	px = pos[G.target(e)];
      }
      if (px.sqr_dist(pos[t]) <= p.sqr_dist(pos[t])) {
	entry_count = count;
	POINT pold = p;
	p = px;
#ifdef DBGWIN
	W.draw_filled_node(p, orange);
	W.draw_filled_node(pold, orange);
#endif // DBGWIN
	entry = e;
      }
      eold = e;
      e = G.face_cycle_succ(e);
    } while (e != start);

    // sanity check
    if (entry == nil) {
      cerr << "KSU_PATH: entry == nil" << endl;
      cout << pos[s] << endl << pos[t] << endl;
      P.clear();
      return;
    }

    // return to the entry edge and cross over into the next face
    if (entry_count > count / 2) {
      e = G.face_cycle_pred(e);
      while (e != entry) {
	P.append(G.reversal(e));
	eold = e;
	e = G.face_cycle_pred(e);
#ifdef DBGWIN
	W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
#endif // DBGWIN
      }
      e = G.reversal(e);
    } else {
      while (e != entry) {
#ifdef DBGWIN
	W.draw_segment(pos[G.source(e)], pos[G.target(e)], orange);
	W.draw_segment(pos[G.source(eold)], pos[G.target(eold)], orange);
#endif // DBGWIN
	P.append(e);
	eold = e;
	e = G.face_cycle_succ(e);
      }
      e = G.cyclic_adj_pred(e);
    }
  }
}



