//==========================================================================
// File: unitgraph.c
// Description: Utilities for building unit graphs
// Author: Pat Morin
// Created: Mon Mar 29 19:09:25 EST 1999
//==========================================================================
#include "unitgraph.h"

#include <stdlib.h>
#include <math.h>

#include <LEDA/ray.h>

// A sleazy parameter passing trick
static GRAPH<point, int> *pG;

//==========================================================================
// The comparison function for sorting edges --- NOT MT SAFE.
//========================================================================
static int slope_cmp(const edge &e1, const edge &e2)
{
  double dir1 = ray((*pG)[pG->source(e1)], (*pG)[pG->target(e1)]).direction();
  double dir2 = ray((*pG)[pG->source(e2)], (*pG)[pG->target(e2)]).direction();
  double result = dir1 - dir2;
  return (result < 0) ? -1 : (result > 0);
}

//==========================================================================
// Build a unit graph G of radius r from the points in L.
//========================================================================
void unit_graph(list<point> &L, double r, GRAPH<point,int> &G)
{
  // Get rid of any crap that might be in G.
  G.clear();

  // Create all the nodes
  point p;
  forall (p, L) {
    G[G.new_node()] = p;
  }

  // Add all edges of length less than r
  node u, v;
  forall_nodes (u, G) {
    for (v = G.succ_node(u); v != nil; v = G.succ_node(v)) {
      if (G[u].distance(G[v]) <= r) {
	G.set_reversal(G.new_edge(u, v, 0), G.new_edge(v, u, 0));
      }
    }
  }

  // sort the edges in counterclockwise order about each node
  pG = &G;
  G.sort_edges(slope_cmp);
  pG = NULL;
}

//==========================================================================
// Build the unit graph G with nedges edges on the points of L.
//========================================================================
void unit_graph(list<point> &L, int nedges, GRAPH<point,int> &G)
{
  // determine the radius that will give us nedges edges
  list_item it1, it2;
  double r;
  {  
    array<double> d((L.size()*(L.size()-1))/2);
    int i = 0;
    forall_items (it1, L) {
      for (it2 = L.succ(it1); it2 != nil; it2 = L.succ(it2)) {
	d[i++] = L.inf(it1).distance(L.inf(it2));
      }
    }
    d.sort();
    r = d[nedges-1];
  }
  // build a graph of this radius
  unit_graph(L, r, G);
}

